/*
 * Copyright (C) 2011-2012 by Matthias Ringwald
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 * 4. This software may not be used in a commercial product
 *    without an explicit license granted by the copyright holder. 
 *
 * THIS SOFTWARE IS PROVIDED BY MATTHIAS RINGWALD AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL MATTHIAS
 * RINGWALD OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

//*****************************************************************************
//
// att device demo
//
//*****************************************************************************

// TODO: seperate BR/EDR from LE ACL buffers
// TODO: move LE init into HCI
// ..

// NOTE: Supports only a single connection

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "global.h"

#include <btstack/run_loop.h>

#include "btstack_memory.h"
#include "hci.h"
#include "hci_dump.h"
#include "l2cap.h"
#include "btstack/debug.h"

#include "att.h"

hci_transport_t * hci_transport_picusb_instance();

//Advertise data
const uint8_t adv_data[31]=
	"\x02\x01\x05"
	"\x0b\x09""FirmaSBBLE"
	"\x03\x02\xf0\xff"
	;

static att_connection_t att_connection;
static uint16_t         att_response_handle = 0;
static uint16_t         att_response_size   = 0;
static uint8_t          att_response_buffer[28];

uint8_t	connection_status=0;

static void on_connect(void);	//sibu
static void on_disconnect(void);	//sibu

void hexdump2(void *data, int size){
    int i;
    for (i=0; i<size;i++){
        log_info("%02X ", ((uint8_t *)data)[i]);
    }
    log_info("\n");
}

static void att_try_respond(void){
    if (!att_response_size) return;
    if (!att_response_handle) return;
    if (!hci_can_send_packet_now(HCI_ACL_DATA_PACKET)) return;
    
    // update state before sending packet
    uint16_t size = att_response_size;
    att_response_size = 0;
    l2cap_send_connectionless(att_response_handle, L2CAP_CID_ATTRIBUTE_PROTOCOL, att_response_buffer, size);
}


static void att_packet_handler(uint8_t packet_type, uint16_t handle, uint8_t *packet, uint16_t size){
    if (packet_type != ATT_DATA_PACKET) return;
    
    att_response_handle = handle;
    att_response_size = att_handle_request(&att_connection, packet, size, att_response_buffer);
    att_try_respond();
}


// enable LE, setup ADV data
static void packet_handler (void * connection, uint8_t packet_type, uint16_t channel, uint8_t *packet, uint16_t size){
    static bd_addr_t addr;
    switch (packet_type) {
            
		case HCI_EVENT_PACKET:
			switch (packet[0]) {
				
                case BTSTACK_EVENT_STATE:
					// bt stack activated, get started - set local name
					if (packet[2] == HCI_STATE_WORKING) {
					   log_info("Working!\n");
						hci_send_cmd(&hci_read_local_supported_features);
					}
					break;
                
                case DAEMON_EVENT_HCI_PACKET_SENT:
                    att_try_respond();
                    break;
                    
                case HCI_EVENT_LE_META:
                    switch (packet[2]) {
                        case HCI_SUBEVENT_LE_CONNECTION_COMPLETE:
                            // reset connection MTU
                            att_connection.mtu = 23;
                            break;
                        default:
                            break;
                    }
                    break;

                case BTSTACK_EVENT_NR_CONNECTIONS_CHANGED:
				    if (packet[2]) {
						on_connect();
                        log_info("CONNECTED\n");
                    } else {
						on_disconnect();
				        log_info("NOT CONNECTED\n");
                    }
                    break;
                    
                case HCI_EVENT_DISCONNECTION_COMPLETE:
                    att_response_handle =0;
                    att_response_size = 0;
                    
                    // restart advertising
                    hci_send_cmd(&hci_le_set_advertise_enable, 1);
                    break;
                    
				case HCI_EVENT_COMMAND_COMPLETE:
					if (COMMAND_COMPLETE_EVENT(packet, hci_read_bd_addr)){
    					bt_flip_addr(addr, &packet[6]);
					    log_info("BD ADDR: %s\n", bd_addr_to_str(addr));
						break;
					}
					if (COMMAND_COMPLETE_EVENT(packet, hci_read_local_supported_features)){
					    log_info("Local supported features: %04lX%04lX\n", READ_BT_32(packet, 10), READ_BT_32(packet, 6));
                        hci_send_cmd(&hci_set_event_mask, 0xffffffff, 0x20001fff);
						break;
					}
					if (COMMAND_COMPLETE_EVENT(packet, hci_set_event_mask)){
                        hci_send_cmd(&hci_write_le_host_supported, 1, 1);
						break;
					}
					if (COMMAND_COMPLETE_EVENT(packet, hci_write_le_host_supported)){
				        hci_send_cmd(&hci_le_set_event_mask, 0xffffffff, 0xffffffff);
                    	break;
					}
					if (COMMAND_COMPLETE_EVENT(packet, hci_le_set_event_mask)){
				        hci_send_cmd(&hci_le_read_buffer_size);
                    	break;
					}
				    if (COMMAND_COMPLETE_EVENT(packet, hci_le_read_buffer_size)){
					    log_info("LE buffer size: %u, count %u\n", READ_BT_16(packet,6), packet[8]);
					   hci_send_cmd(&hci_le_read_supported_states);
					   break;
					}
				    if (COMMAND_COMPLETE_EVENT(packet, hci_le_read_supported_states)){
					   hci_send_cmd(&hci_le_set_advertising_parameters,  0x0400, 0x0800, 0, 0, 0, &addr, 0x07, 0);
					   break;
					}
				    if (COMMAND_COMPLETE_EVENT(packet, hci_le_set_advertising_parameters)){
					   hci_send_cmd(&hci_le_set_advertising_data, sizeof(adv_data), adv_data);
					   break;
					}
				    if (COMMAND_COMPLETE_EVENT(packet, hci_le_set_advertising_data)){
					   hci_send_cmd(&hci_le_set_scan_response_data, 10, adv_data);
					   break;
					}
				    if (COMMAND_COMPLETE_EVENT(packet, hci_le_set_scan_response_data)){
					   hci_send_cmd(&hci_le_set_advertise_enable, 1);
					   break;
					}
				    if (COMMAND_COMPLETE_EVENT(packet, hci_le_set_advertise_enable)){
                        hci_discoverable_control(1);
						log_info("startup_state=1\n");
						startup_state=1;
					    led1_off();
                        break;
					}
                    
			}
	}
}

// gatt profile
#include "profile.h"
#include "attribute.h"

#define MODE_INPUT    0x00
#define MODE_OUTPUT   0x01
#define MODE_ANALOG   0x02
#define MODE_PWM      0x03
#define MODE_SERVO    0x04
#define MODE_SHIFT    0x05
#define MODE_I2C      0x06

#define START_SYSEX             0xF0 // start a MIDI Sysex message
#define END_SYSEX               0xF7 // end a MIDI Sysex message
#define PIN_MODE_QUERY          0x72 // ask for current and supported pin modes
#define PIN_MODE_RESPONSE       0x73 // reply with current and supported pin modes
#define PIN_STATE_QUERY         0x6D
#define PIN_STATE_RESPONSE      0x6E
#define CAPABILITY_QUERY        0x6B
#define CAPABILITY_RESPONSE     0x6C
#define ANALOG_MAPPING_QUERY    0x69
#define ANALOG_MAPPING_RESPONSE 0x6A
#define REPORT_FIRMWARE         0x79 // report name and version of the firmware
#define	SYSEX_I2C_REPLY			0x77

#define	FLAG_NOTIFICATION			0x01

#define	FUNC_DIN					0x01
#define	FUNC_DOUT					0x02
#define	FUNC_SERVO					0x04
#define	FUNC_PWM					0x08
#define	FUNC_AIN					0x10
#define	FUNC_AOUT					0x20

typedef struct{
	uint8_t		pinno;
	uint8_t		enable;
	uint8_t		func;
} PORTMAP;

/*const*/ static PORTMAP portmap[]={
	{ 2,1,	FUNC_DIN|FUNC_DOUT|FUNC_SERVO,},
	{ 3,1,	FUNC_DIN|FUNC_DOUT|FUNC_SERVO,},
	{ 4,1,	FUNC_DIN|FUNC_DOUT|FUNC_SERVO,},
	{ 5,1,	FUNC_DIN|FUNC_DOUT|FUNC_SERVO,},
	{ 6,1,	FUNC_DIN|FUNC_DOUT|FUNC_SERVO,},
	{ 7,1,	FUNC_DIN|FUNC_DOUT|FUNC_SERVO,},
	{ 8,1,	FUNC_DIN|FUNC_DOUT|FUNC_SERVO,},
	{ 9,1,	FUNC_DIN|FUNC_DOUT|FUNC_SERVO,},

	{14,1,	FUNC_AIN,},
	{15,1,	FUNC_AIN,},
	{16,1,	FUNC_AIN,},
	{17,1,	FUNC_AIN,},
	{18,1,	FUNC_AIN,},
	{19,1,	FUNC_AIN,},
};

#define	PORTMAPS		(sizeof(portmap)/sizeof(portmap[0]))

static uint8_t	pio_setting=0;
static uint8_t	pio_pullup=0;
static uint8_t	pio_input=0;
static uint8_t	pio_output=0;
//static uint8_t	uart_config=0;
//static uint16_t	uart_baudrate=UART_DEFAULT_BAUDRATE/240;
//static uint8_t	uart_lasttx_size=0;
//static uint8_t	uart_lasttx_str[20];
static uint32_t	pwm_period[8];
static uint32_t	pwm_duty[8];
static uint8_t	pwm_config;
uint8_t			uart1_polarity;
static uint8_t	i2c_recvbuf[20];
static uint8_t	i2c_recvlen;

static uint16_t	din_state;
static uint16_t	ain_state[8];

//characteristic config
static uint8_t	pio_input_config=0;
static uint8_t	uart_rx_config=0;
static uint8_t	lowbat_notify_config=0;

#define	ADIN_VREF_VALUE	5000
#define	ADIN_FULLSCALE	1023

uint16_t	advalue2mv(uint16_t advalue)
{
	uint16_t	ret;

	ret=(uint16_t)((((uint32_t)advalue)*ADIN_VREF_VALUE)/ADIN_FULLSCALE);
	return ret;
}

void UART1PutBuf(uint8_t *buf,uint8_t bufsize)
{
	if(!bufsize){
		return;
	}
	for(;bufsize--;buf++){
		UART1PutChar(*buf);
	}
	return;
}

void gpio_setup(uint8_t setup)
{
	int			mapno;
	int			output;
	uint8_t		buf[3];

	for(mapno=0;mapno<PORTMAPS;mapno++){
		if(!portmap[mapno].enable){
			continue;
		}
		output=setup&(1<<mapno);

		if(output){
			if(!(portmap[mapno].func&FUNC_DOUT)){
				continue;
			}
		}else{
			if(!(portmap[mapno].func&FUNC_DIN)){
				continue;
			}
		}
		buf[0]=0xf4;
		buf[1]=portmap[mapno].pinno;
		buf[2]=output?MODE_OUTPUT:MODE_INPUT;	
		UART1PutBuf(buf,sizeof(buf));
	}
	return;
}

uint8_t gpio_input(void)
{
	uint8_t		data=0;
	unsigned	bitmask;
	int			mapno;

	for(mapno=0;mapno<PORTMAPS;mapno++){
		if(!portmap[mapno].enable){
			continue;					//ɂȂĂ
		}
		bitmask=(1<<mapno);
		if(pio_setting&bitmask){
			continue;					//o͐ݒɂȂĂ
		}
		if(pwm_config&bitmask){
			continue;					//PWMo͂ɂȂĂ
		}

		if(din_state&(1<<portmap[mapno].pinno)){
			data|=bitmask;
		}
	}

	return data;
}

void gpio_output(uint8_t data)
{
	int			mapno;
	unsigned	bitmask;
	uint8_t		buf[3];
	uint16_t	port_val;

	memset(buf,0,sizeof(buf));
	port_val=0;
	for(mapno=0;mapno<PORTMAPS;mapno++){
		if(!portmap[mapno].enable){
			continue;
		}
		bitmask=(1<<mapno);
		if(!(pio_setting&bitmask)){
			continue;					//͐ݒɂȂĂ
		}
		if(pwm_config&bitmask){
			continue;					//PWMo͂ɂȂĂ
		}
		if(!(portmap[mapno].func&FUNC_DOUT)){
			continue;
		}
		if(data&bitmask){
			port_val|=(1<<portmap[mapno].pinno);	//sԍ15܂ł̑Ή
		}
	}
	buf[0]=0x90;
	buf[1]=(port_val>> 0)&0x7f;
	buf[2]=(port_val>> 7)&0x01;
	UART1PutBuf(buf,sizeof(buf));

	buf[0]=0x91;
	buf[1]=(port_val>> 8)&0x7f;
	buf[2]=(port_val>>15)&0x01;
	UART1PutBuf(buf,sizeof(buf));

	return;
}

void gpio_setpullup(uint8_t pullupsetting)
{
	return;
}

#if	0
static void pwm_setup(uint8_t setup)
{
	int			mapno;
	int			output;
	uint8_t		buf[3];

	for(mapno=0;mapno<PORTMAPS;mapno++){
		if(!portmap[mapno].enable){
			continue;
		}
		output=setup&(1<<mapno);

		if(!output || !(portmap[mapno].func&FUNC_PWM)){
			continue;
		}

		buf[0]=0xf4;
		buf[1]=portmap[mapno].pinno;
		buf[2]=MODE_PWM;	
		UART1PutBuf(buf,sizeof(buf));
	}
	return;
}
#endif

#define	us2count(x)		((x)/PWM_RATIO_US)

static void set_pwm(int ch)
{
#if	0
	uint32_t	period,duty;

	if(ch>=5){
		return;
	}
	period=us2count(pwm_period[ch]);
	duty=us2count(pwm_duty[ch]);

	if(period<2){
		period=2;
	}else
	if(period>65536){
		period=65536;
	}
	period-=1;

	if(duty>65535){
		duty=65535;
	}

	switch(ch){
	case	0:
		SetDCOC1PWM(duty,period);
		break;
	case	1:
		SetDCOC2PWM(duty,period);
		break;
	case	2:
		SetDCOC3PWM(duty,period);
		break;
	case	3:
		SetDCOC4PWM(duty,period);
		break;
	case	4:
		SetDCOC5PWM(duty,period);
		break;
	default:
		break;
	}
	return;
#endif
}

static void set_pwm_config(uint8_t config)
{
#if	0
	static uint8_t	last_config=0;
//	uint16_t		width,duty,i;

	if(!last_config && config){
		//initialize pwm
		OpenTimer2(T2_ON|PWM_PRESCALER,PWM_TIMER_PERIOD-1);
	}else
	if(last_config && !config){
		//terminate pwm
		CloseOC1();
		CloseOC2();
		CloseOC3();
		CloseOC4();
		CloseOC5();
		CloseTimer2();
	}

	if(!(last_config&0x01) && config&0x01){
		OpenOC1(
			OC_IDLE_CON|OC_TIMER2_SRC|OC_PWM_EDGE_ALIGN,
			OC_TRIGGER_ENABLE|OC_SYNC_TRIG_IN_TMR2,
			0,
			0
		);
		set_pwm(0);
	}else
	if(last_config&0x01 && !(config&0x01)){
		SetDCOC1PWM(0,0);
		CloseOC1();
	}

	if(!(last_config&0x02) && config&0x02){
		OpenOC2(
			OC_IDLE_CON|OC_TIMER2_SRC|OC_PWM_EDGE_ALIGN,
			OC_TRIGGER_ENABLE|OC_SYNC_TRIG_IN_TMR2,
			0,
			0
		);
		set_pwm(1);
	}else
	if(last_config&0x02 && !(config&0x02)){
		SetDCOC2PWM(0,0);
		CloseOC2();
	}

	if(!(last_config&0x04) && config&0x04){
		OpenOC3(
			OC_IDLE_CON|OC_TIMER2_SRC|OC_PWM_EDGE_ALIGN,
			OC_TRIGGER_ENABLE|OC_SYNC_TRIG_IN_TMR2,
			0,
			0
		);
		set_pwm(2);
	}else
	if(last_config&0x04 && !(config&0x04)){
		SetDCOC3PWM(0,0);
		CloseOC3();
	}

	if(!(last_config&0x08) && config&0x08){
		OpenOC4(
			OC_IDLE_CON|OC_TIMER2_SRC|OC_PWM_EDGE_ALIGN,
			OC_TRIGGER_ENABLE|OC_SYNC_TRIG_IN_TMR2,
			0,
			0
		);
		set_pwm(3);
	}else
	if(last_config&0x08 && !(config&0x08)){
		SetDCOC4PWM(0,0);
		CloseOC4();
	}

	if(!(last_config&0x10) && config&0x10){
		OpenOC5(
			OC_IDLE_CON|OC_TIMER2_SRC|OC_PWM_EDGE_ALIGN,
			OC_TRIGGER_ENABLE|OC_SYNC_TRIG_IN_TMR2,
			0,
			0
		);
		set_pwm(4);
	}else
	if(last_config&0x10 && !(config&0x10)){
		SetDCOC5PWM(0,0);
		CloseOC5();
	}

	pwm_setup(config);
	last_config=config;

#endif
	return;
}

static void init_firmata(void)
{
	unsigned	i,mapno;
	uint8_t		mode;
	uint8_t		buf[2];

	for (i=0; i<16; i++) {
		buf[0] = 0xC0 | i;  // report analog
		buf[1] = 1;
		UART1PutBuf(buf,sizeof(buf));

		buf[0] = 0xD0 | i;  // report digital
		buf[1] = 1;
		UART1PutBuf(buf,sizeof(buf));
	}

	//l=fW^
	for(mapno=0;mapno<PORTMAPS;mapno++){
		if(!portmap[mapno].enable){
			continue;
		}
		if((portmap[mapno].func&FUNC_AIN)){
			mode=MODE_ANALOG;
		}else{
			mode=MODE_INPUT;
		}
		buf[0]=0xf4;
		buf[1]=portmap[mapno].pinno;
		buf[2]=mode;	
		UART1PutBuf(buf,sizeof(buf));
	}

	return;
}

static void on_connect(void)	//sibu
{
	connection_status=1;

	init_firmata();
	return;
}

static void on_disconnect(void)	//sibu
{
	connection_status=0;

	pio_input_config=
	uart_rx_config=
	lowbat_notify_config=
		0;

	return;
}

// read requests
static uint16_t att_read_callback(uint16_t handle, uint16_t offset, uint8_t * buffer, uint16_t buffer_size){
	uint16_t	ret=0,i;
	//int			c;

	if(buffer){
		log_info("READ Callback, handle %04x\n", handle);
		led1_on_mac();
	}
	switch(handle){
	case	HANDLE_PIO_SETTING_VALUE:
		ret=1;
		if(buffer && buffer_size){
			buffer[0]=pio_setting;
		}
		break;
	case	HANDLE_PIO_PULLUP_VALUE:
		ret=1;
		if(buffer && buffer_size){
			buffer[0]=pio_pullup;
		}
		break;
	case	HANDLE_PIO_INPUT_VALUE:
		ret=1;
		if(buffer && buffer_size){
			buffer[0]=pio_input;
		}
		break;
	case	HANDLE_AI_READ0_VALUE:
		ret=2;
		if(buffer && buffer_size>=2){
			i=advalue2mv(ain_state[0]);
			buffer[0]=i>>8;
			buffer[1]=i&0xff;
		}
		break;
	case	HANDLE_AI_READ1_VALUE:
		ret=2;
		if(buffer && buffer_size>=2){
			i=advalue2mv(ain_state[1]);
			buffer[0]=i>>8;
			buffer[1]=i&0xff;
		}
		break;

	case	HANDLE_AI_READ2_VALUE:
		ret=2;
		if(buffer && buffer_size>=2){
			i=advalue2mv(ain_state[2]);
			buffer[0]=i>>8;
			buffer[1]=i&0xff;
		}
		break;

	case	HANDLE_I2C_READ_VALUE:
		//received bytes
		ret=i2c_recvlen;
		if(buffer){
			if(buffer_size>=i2c_recvlen){
				for(i=0;i<i2c_recvlen;i++){
					buffer[i]=i2c_recvbuf[i];
				}
				for(;i<buffer_size;i++){
					buffer[i]=0;	//padding
				}
			}
			i2c_recvlen=0;
		}
		break;
#if	0
	case	HANDLE_UART_CONFIG_VALUE:
		ret=1;
		if(buffer && buffer_size){
			buffer[0]=uart_config;
		}
		break;
	case	HANDLE_UART_BAUDRATE_VALUE:
		ret=2;
		if(buffer && buffer_size){
			buffer[0]=uart_baudrate>>8;
			buffer[1]=uart_baudrate&0xff;
		}
		break;
	case	HANDLE_UART_TX_VALUE:
		ret=uart_lasttx_size;
		if(buffer && buffer_size){
			for(i=0;i<buffer_size;i++){
				if(i<uart_lasttx_size){
					buffer[i]=uart_lasttx_str[i];
				}else{
					buffer[i]=0;	//padding
				}
			}
		}
		break;
	case	HANDLE_UART_RX_VALUE:
		ret=UART1GetRXCnt();
#if	UART_PACKET_MULTIPLE_BYTES
		if(ret>=(att_connection.mtu-3)){
			ret=(att_connection.mtu-3);
		}
#else
		if(ret){
			ret=1;
		}
#endif
		if(buffer && buffer_size){
            log_info("UART1 RX: ");
			for(i=0;i<buffer_size;i++){
				if(i<ret){
					c=UART1GetChar();
					if(c<0){
						c=0;			//padding
					}
				}else{
					c=0;				//padding
				}
				buffer[i]=(uint8_t)c;
			}
		}
		break;
#endif
	case	HANDLE_LOWBAT_NOTIFY_VALUE:
		ret=1;
		if(buffer && buffer_size){
			buffer[0]=0;		//no lowbat
		}
		break;
	case	HANDLE_PIO_INPUT_CONFIG:
		ret=1;
		if(buffer && buffer_size){
			buffer[0]=pio_input_config;
		}
		break;
	case	HANDLE_UART_RX_CONFIG:
		ret=1;
		if(buffer && buffer_size){
			buffer[0]=uart_rx_config;
		}
		break;
	case	HANDLE_LOWBAT_NOTIFY_CONFIG:
		ret=1;
		if(buffer && buffer_size){
			buffer[0]=lowbat_notify_config;
		}
		break;
	case	HANDLE_MACHINE_ID_VALUE:
		ret=4;
		if(buffer && buffer_size>=4){
			net_store_32(buffer,0,MACHINE_ID);
		}
		break;
	case	HANDLE_UART_POLARITY_VALUE:
		ret=1;
		if(buffer && buffer_size){
			buffer[0]=uart1_polarity;
		}
		break;
	default:
		break;
    }
	return ret;
}

uint16_t firmata_byte2word(uint8_t c)
{
	uint16_t	ret;

	ret=((uint16_t)(c&0x80))<<1;
	ret|=(uint16_t)(c&0x7f);

	return ret;
}

uint8_t firmata_word2byte(uint8_t c1,uint8_t c2)
{
	uint8_t	ret;

	ret=(c2<<7)|(c1);
	return ret;
}

void i2c_write(uint8_t *buffer,uint16_t buffer_size)
{
	uint8_t		i,len;
	uint16_t	w;

	UART1PutChar(START_SYSEX);
	UART1PutChar(0x76);

	len=buffer[0]-1;

	//slave address
	UART1PutChar(buffer[1]>>1);
	UART1PutChar(0x00);

	for(i=0;i<len;i++){
		w=firmata_byte2word(buffer[2+i]);
		UART1PutChar((uint8_t)(w&0xff));
		UART1PutChar((uint8_t)(w>>8));
	}
	UART1PutChar(END_SYSEX);
}

void i2c_read(uint8_t *buffer,uint16_t buffer_size)
{
	uint8_t		len;

	UART1PutChar(START_SYSEX);
	UART1PutChar(0x76);

	len=buffer[0];

	UART1PutChar(buffer[1]>>1);
	UART1PutChar(0x08);

	UART1PutChar(len&0x7f);
	UART1PutChar(len>>7);

	UART1PutChar(END_SYSEX);
}

// write requests
static void att_write_callback(uint16_t handle, uint16_t transaction_mode, uint16_t offset, uint8_t *buffer, uint16_t buffer_size, signature_t * signature){
	uint16_t	i;
	uint32_t	l;

    log_info("WRITE Callback, handle %04x\n", handle);
	led1_on_mac();
    switch(handle){
	case	HANDLE_PIO_SETTING_VALUE:
		if(buffer_size){
			pio_setting=buffer[0];
			gpio_setup(pio_setting);
		}
		break;
	case	HANDLE_PIO_PULLUP_VALUE:
		if(buffer_size){
			pio_pullup=buffer[0];
			gpio_setpullup(pio_pullup);
		}
		break;
	case	HANDLE_PIO_OUTPUT_VALUE:
		if(buffer_size){
			pio_output=buffer[0];
			gpio_output(buffer[0]);
		}
		break;
	case	HANDLE_PWM_CONFIG_VALUE:
		if(buffer_size){
			pwm_config=buffer[0];
			set_pwm_config(pwm_config);
		}
		break;
	case	HANDLE_PWM_PARAM_VALUE:
		if(buffer_size>=5){
			i=buffer[0];
			l =((uint32_t)buffer[1])<<24;
			l|=((uint32_t)buffer[2])<<16;
			l|=((uint32_t)buffer[3])<< 8;
			l|=((uint32_t)buffer[4])<< 0;
			if(i<8){
				pwm_period[i]=l;
				set_pwm(i);
			}
		}
		break;
	case	HANDLE_PWM_DUTY_VALUE:
		if(buffer_size>=5){
			i=buffer[0];
			l =((uint32_t)buffer[1])<<24;
			l|=((uint32_t)buffer[2])<<16;
			l|=((uint32_t)buffer[3])<< 8;
			l|=((uint32_t)buffer[4])<< 0;
			if(i<8){
				pwm_duty[i]=l;
				set_pwm(i);
			}
		}
		break;
	case	HANDLE_I2C_CONFIG_VALUE:
		//write 0:OFF 1:100kHz 2:400kHz
		UART1PutChar(START_SYSEX);
		UART1PutChar(0x78);
		UART1PutChar(0x00);
		UART1PutChar(0x00);
		UART1PutChar(END_SYSEX);
		break;
	case	HANDLE_I2C_START_STOP_VALUE:
		//write 0:stop 1:start condition
		break;
	case	HANDLE_I2C_WRITE_VALUE:	//write command
		//write +0:0x00 +1:write length(from next byte) +2:(addr<<1)|(R:1/W:0) +3`:Write bytes
		if(buffer_size>=3){
			i2c_write(buffer,buffer_size);
		}
		break;
	case	HANDLE_I2C_READ_PARAM_VALUE:	//read command
		//write +0:0x00 +1:read length(from next byte) +2:(addr<<1)|(R:1/W:0)
		if(buffer_size>=2){
			i2c_read(buffer,buffer_size);
		}
		break;

#if	0
	case	HANDLE_UART_CONFIG_VALUE:
		if(buffer_size){
			uart_config=buffer[0];
			if(uart_config){
				//UART̃|[ggpioŎgp֎~ɂ
				portmap[UART_MAPNO1].enable=0;
				portmap[UART_MAPNO2].enable=0;
#if	UART_FIX_BAUDRATE
				UART1Init(UART_DEFAULT_BAUDRATE);
#else
				UART1Init(uart_baudrate*240LU);
#endif
			}else{
				UART1Term();
				//UART̃|[ggpioŎgp\ɂ
				portmap[UART_MAPNO1].enable=1;
				portmap[UART_MAPNO2].enable=1;
			}
		}
		break;
	case	HANDLE_UART_BAUDRATE_VALUE:
		if(buffer_size>=2){
			uart_baudrate=((uint16_t)(buffer[0]))<<8;
			uart_baudrate+=buffer[1];
		}
		break;
	case	HANDLE_UART_TX_VALUE:
		if(buffer_size){
            log_info("UART1 TX: ");
			for(i=0;i<buffer_size;i++){
				log_info("%02x ",buffer[i]);
				UART1PutChar(buffer[i]);
			}
			log_info("\n");
		}
		break;
#endif
	case	HANDLE_PIO_INPUT_CONFIG:
		if(buffer_size){
			pio_input_config=buffer[0];
		}
		break;
	case	HANDLE_UART_RX_CONFIG:
		if(buffer_size){
			uart_rx_config=buffer[0];
		}
		break;
	case	HANDLE_HARDWARE_RESET_VALUE:
		if(buffer_size){
			if(buffer[0]){
				Reset();
			}
		}
		break;
	case	HANDLE_LOWBAT_NOTIFY_CONFIG:
		if(buffer_size){
			lowbat_notify_config=buffer[0];
		}
		break;
	case	HANDLE_UART_POLARITY_VALUE:
		if(buffer_size){
			uart1_polarity=buffer[0];
		}
		break;
    }
}

enum{
STATE_IDLE=0,
STATE_WAIT,
STATE_CMD,
};

void	firmata_receive_process(void)
{
	static uint8_t	recv_buf[48];
	static uint8_t	recv_len;
	static uint8_t	wait_len;
	static uint8_t	state;
	int				c;
	uint16_t		u,len;

	for(;;){
		switch(state){
		case	STATE_IDLE:
			c=UART1GetChar();
			if(c<0){
				break;
			}
			if(
				(c>=0x90 && c<=0x91)
			||	(c>=0xe0 && c<=0xe7)
			){
				wait_len=3;
			}else
			if(c==START_SYSEX){
				//0xf0
				//  find stop mark 0xf7
				wait_len=0;
			}else{
				continue;
			}
			recv_buf[0]=(uint8_t)c;
			recv_len=1;
			state=STATE_WAIT;
			continue;

		case	STATE_WAIT:
			c=UART1GetChar();
			if(c<0){
				break;
			}
			recv_buf[recv_len++]=(uint8_t)c;
			if(wait_len){
				if(recv_len>=wait_len){
					state=STATE_CMD;
				}
			}else{
				//sysex find 0xf7
				if(c==END_SYSEX){
					state=STATE_CMD;
				}
			}
			continue;

		case	STATE_CMD:
			switch(recv_buf[0]){
			case	START_SYSEX:
				switch(recv_buf[1]){
				case	SYSEX_I2C_REPLY:
					//recv_buf[2-3]:slave address
					//recv_buf[4-5]:register address
					//recv_buf[6-7]:data[0]
					//recv_buf[8-9]:data[1]...
					//log_info("SYSEX I2C REPLY in\n");
					if(!i2c_recvlen && recv_len>6){
						len=(recv_len-6)/2;
						for(u=0;u<len;u++){
							i2c_recvbuf[u]=firmata_word2byte(recv_buf[6+u*2+0],recv_buf[6+u*2+1]);
						}
						i2c_recvlen=len;
						//log_info("SYSEX I2C REPLY:%d bytes\n",len);
					}
					break;
				}
				break;
			case	0x90:
				u=din_state&0xff00;
				u|=((recv_buf[2]&1)<<7)|((recv_buf[1]&0x7f)<<0);
				din_state=u;
				break;
			case	0x91:
				u=din_state&0x00ff;
				u|=((recv_buf[2]&1)<<15)|((recv_buf[1]&0x7f)<<8);
				din_state=u;
				break;
			case	0xe0:
				ain_state[0]=((uint16_t)recv_buf[2]<<7)|((uint16_t)recv_buf[1]);
				break;
			case	0xe1:
				ain_state[1]=((uint16_t)recv_buf[2]<<7)|((uint16_t)recv_buf[1]);
				break;
			case	0xe2:
				ain_state[2]=((uint16_t)recv_buf[2]<<7)|((uint16_t)recv_buf[1]);
				break;
			case	0xe3:
				ain_state[3]=((uint16_t)recv_buf[2]<<7)|((uint16_t)recv_buf[1]);
				break;
			case	0xe4:
				ain_state[4]=((uint16_t)recv_buf[2]<<7)|((uint16_t)recv_buf[1]);
				break;
			case	0xe5:
				ain_state[5]=((uint16_t)recv_buf[2]<<7)|((uint16_t)recv_buf[1]);
				break;
			case	0xe6:
				ain_state[6]=((uint16_t)recv_buf[2]<<7)|((uint16_t)recv_buf[1]);
				break;
			case	0xe7:
				ain_state[7]=((uint16_t)recv_buf[2]<<7)|((uint16_t)recv_buf[1]);
				break;
			}
			state=STATE_IDLE;
			continue;
		}
		break;
	}
	return;
}


// check notification
void check_notification(void)
{
	int i,j,c;
	uint8_t	notify_buf[20];
	static uint8_t	pio_last_input=0;

	firmata_receive_process();
	if(connection_status){
		if(
			uart_rx_config&FLAG_NOTIFICATION
		&&	UART1GetRXCnt()
		&&	hci_can_send_packet_now(HCI_ACL_DATA_PACKET)		//M]T͂邩
		){
#if	UART_PACKET_MULTIPLE_BYTES
			j=att_connection.mtu-3;
#else
			j=1;
#endif
			for(i=0;i<j;i++){
				c=UART1GetChar();
				if(c<0){
					break;
				}
				notify_buf[i]=(uint8_t)c;
			}
			if(i){
				log_info("UART RX NOTIFICATION\n");
				att_response_size=att_prepare_handle_value_notification(&att_connection,HANDLE_UART_RX_VALUE,notify_buf,i,att_response_buffer);
				att_try_respond();
			}
		}

		pio_input=gpio_input();
		if(pio_input_config&FLAG_NOTIFICATION){
			if(
				pio_input!=pio_last_input
			&&	hci_can_send_packet_now(HCI_ACL_DATA_PACKET)		//M]T͂邩
			){
				log_info("PIO INPUT NOTIFICATION\n");
				notify_buf[0]=pio_input;
				att_response_size=att_prepare_handle_value_notification(&att_connection,HANDLE_PIO_INPUT_VALUE,notify_buf,1,att_response_buffer);
				att_try_respond();
				pio_last_input=pio_input;
			}
		}else{
			pio_last_input=pio_input;
		}

		if(lowbat_notify_config&FLAG_NOTIFICATION){
		}
	}
}

// main
int blemain(void)
{
	/// GET STARTED with BTstack ///
	btstack_memory_init();
    run_loop_init(RUN_LOOP_EMBEDDED);
	
    // init HCI
	hci_transport_t    * transport = hci_transport_picusb_instance();
	bt_control_t       * control   = NULL;
    hci_uart_config_t  * config    = NULL;
    remote_device_db_t * remote_db = (remote_device_db_t *) &remote_device_db_memory;
	hci_init(transport, config, control, remote_db);
	
    // use eHCILL
    // bt_control_cc256x_enable_ehcill(1);
    
    // set up l2cap_le
    l2cap_init();
    l2cap_register_fixed_channel(att_packet_handler, L2CAP_CID_ATTRIBUTE_PROTOCOL);
    l2cap_register_packet_handler(packet_handler);
    
    // set up ATT
	log_info("profile_data %u bytes.\n",sizeof(profile_data));
    att_set_db(profile_data);
    att_set_write_callback(att_write_callback);
    att_set_read_callback(att_read_callback);
    att_dump_attributes();
    att_connection.mtu = 27;
    
	log_info("Run...\n\r");

    // turn on!
	hci_power_control(HCI_POWER_ON);

    // go!
    run_loop_execute();	
    
    // happy compiler!
    return 0;
}

